@@ -10,8 +10,9 @@
 use std::num::{NonZeroU16, NonZeroU32};
 
 use crate::client::rtp;
+use crate::ConnectionContext;
+use crate::Error;
 use bytes::{Buf, Bytes};
-use failure::{bail, Error};
 use pretty_hex::PrettyHex;
 
 pub(crate) mod aac;
@@ -112,13 +113,7 @@ pub struct AudioParameters {
     frame_length: Option<NonZeroU32>,
     clock_rate: u32,
     extra_data: Bytes,
-    config: AudioCodecConfig,
-}
-
-#[derive(Clone)]
-enum AudioCodecConfig {
-    Aac(aac::AudioSpecificConfig),
-    Other,
+    sample_entry: Option<Bytes>,
 }
 
 impl std::fmt::Debug for AudioParameters {
@@ -151,17 +146,18 @@ impl AudioParameters {
         &self.extra_data
     }
 
-    /// Builds an `.mp4` `SimpleAudioEntry` box (as defined in ISO/IEC 14496-12) if possible.
+    /// An `.mp4` `SimpleAudioEntry` box (as defined in ISO/IEC 14496-12), if possible.
+    ///
     /// Not all codecs can be placed into a `.mp4` file, and even for supported codecs there
     /// may be unsupported edge cases.
-    pub fn sample_entry(&self) -> Result<Bytes, Error> {
-        aac::get_mp4a_box(self)
+    pub fn sample_entry(&self) -> Option<&Bytes> {
+        self.sample_entry.as_ref()
     }
 }
 
 /// An audio frame, which consists of one or more samples.
 pub struct AudioFrame {
-    pub ctx: crate::Context,
+    pub ctx: crate::RtspMessageContext,
     pub stream_id: usize,
     pub timestamp: crate::Timestamp,
     pub frame_length: NonZeroU32,
@@ -206,7 +202,7 @@ impl Buf for AudioFrame {
 pub struct MessageParameters(onvif::CompressionType);
 
 pub struct MessageFrame {
-    pub ctx: crate::Context,
+    pub ctx: crate::RtspMessageContext,
     pub timestamp: crate::Timestamp,
     pub stream_id: usize,
 
@@ -236,7 +232,8 @@ impl std::fmt::Debug for MessageFrame {
 /// specified here; they can be calculated from the timestamp of a following
 /// picture, or approximated via the frame rate.
 pub struct VideoFrame {
-    pub new_parameters: Option<VideoParameters>,
+    // New video parameters. Rarely populated and large, so boxed to reduce bloat.
+    pub new_parameters: Option<Box<VideoParameters>>,
 
     /// Number of lost RTP packets before this video frame. See [crate::client::rtp::Packet::loss].
     /// Note that if loss occurs during a fragmented frame, more than this number of packets' worth
@@ -245,8 +242,8 @@ pub struct VideoFrame {
 
     // A pair of contexts: for the start and for the end.
     // Having both can be useful to measure the total time elapsed while receiving the frame.
-    start_ctx: crate::Context,
-    end_ctx: crate::Context,
+    start_ctx: crate::RtspMessageContext,
+    end_ctx: crate::RtspMessageContext,
 
     /// This picture's timestamp in the time base associated with the stream.
     pub timestamp: crate::Timestamp,
@@ -268,12 +265,12 @@ pub struct VideoFrame {
 
 impl VideoFrame {
     #[inline]
-    pub fn start_ctx(&self) -> crate::Context {
+    pub fn start_ctx(&self) -> crate::RtspMessageContext {
         self.start_ctx
     }
 
     #[inline]
-    pub fn end_ctx(&self) -> crate::Context {
+    pub fn end_ctx(&self) -> crate::RtspMessageContext {
         self.end_ctx
     }
 
@@ -314,10 +311,10 @@ pub struct Depacketizer(DepacketizerInner);
 #[derive(Debug)]
 enum DepacketizerInner {
     Aac(Box<aac::Depacketizer>),
-    SimpleAudio(simple_audio::Depacketizer),
-    G723(g723::Depacketizer),
+    SimpleAudio(Box<simple_audio::Depacketizer>),
+    G723(Box<g723::Depacketizer>),
     H264(Box<h264::Depacketizer>),
-    Onvif(onvif::Depacketizer),
+    Onvif(Box<onvif::Depacketizer>),
 }
 
 impl Depacketizer {
@@ -327,7 +324,7 @@ impl Depacketizer {
         clock_rate: u32,
         channels: Option<NonZeroU16>,
         format_specific_params: Option<&str>,
-    ) -> Result<Self, Error> {
+    ) -> Result<Self, String> {
         use onvif::CompressionType;
 
         // RTP Payload Format Media Types
@@ -342,56 +339,59 @@ impl Depacketizer {
                 channels,
                 format_specific_params,
             )?)),
-            ("audio", "g726-16") => {
-                DepacketizerInner::SimpleAudio(simple_audio::Depacketizer::new(clock_rate, 2))
-            }
-            ("audio", "g726-24") => {
-                DepacketizerInner::SimpleAudio(simple_audio::Depacketizer::new(clock_rate, 3))
-            }
-            ("audio", "dvi4") | ("audio", "g726-32") => {
-                DepacketizerInner::SimpleAudio(simple_audio::Depacketizer::new(clock_rate, 4))
-            }
-            ("audio", "g726-40") => {
-                DepacketizerInner::SimpleAudio(simple_audio::Depacketizer::new(clock_rate, 5))
-            }
+            ("audio", "g726-16") => DepacketizerInner::SimpleAudio(Box::new(
+                simple_audio::Depacketizer::new(clock_rate, 2),
+            )),
+            ("audio", "g726-24") => DepacketizerInner::SimpleAudio(Box::new(
+                simple_audio::Depacketizer::new(clock_rate, 3),
+            )),
+            ("audio", "dvi4") | ("audio", "g726-32") => DepacketizerInner::SimpleAudio(Box::new(
+                simple_audio::Depacketizer::new(clock_rate, 4),
+            )),
+            ("audio", "g726-40") => DepacketizerInner::SimpleAudio(Box::new(
+                simple_audio::Depacketizer::new(clock_rate, 5),
+            )),
             ("audio", "pcma") | ("audio", "pcmu") | ("audio", "u8") | ("audio", "g722") => {
-                DepacketizerInner::SimpleAudio(simple_audio::Depacketizer::new(clock_rate, 8))
-            }
-            ("audio", "l16") => {
-                DepacketizerInner::SimpleAudio(simple_audio::Depacketizer::new(clock_rate, 16))
+                DepacketizerInner::SimpleAudio(Box::new(simple_audio::Depacketizer::new(
+                    clock_rate, 8,
+                )))
             }
+            ("audio", "l16") => DepacketizerInner::SimpleAudio(Box::new(
+                simple_audio::Depacketizer::new(clock_rate, 16),
+            )),
             // Dahua cameras when configured with G723 send packets with a
             // non-standard encoding-name "G723.1" and length 40, which doesn't
             // make sense. Don't try to depacketize these.
-            ("audio", "g723") => DepacketizerInner::G723(g723::Depacketizer::new(clock_rate)?),
-            ("application", "vnd.onvif.metadata") => {
-                DepacketizerInner::Onvif(onvif::Depacketizer::new(CompressionType::Uncompressed))
-            }
-            ("application", "vnd.onvif.metadata.gzip") => {
-                DepacketizerInner::Onvif(onvif::Depacketizer::new(CompressionType::GzipCompressed))
-            }
-            ("application", "vnd.onvif.metadata.exi.onvif") => {
-                DepacketizerInner::Onvif(onvif::Depacketizer::new(CompressionType::ExiDefault))
-            }
-            ("application", "vnd.onvif.metadata.exi.ext") => {
-                DepacketizerInner::Onvif(onvif::Depacketizer::new(CompressionType::ExiInBand))
+            ("audio", "g723") => {
+                DepacketizerInner::G723(Box::new(g723::Depacketizer::new(clock_rate)?))
             }
+            ("application", "vnd.onvif.metadata") => DepacketizerInner::Onvif(Box::new(
+                onvif::Depacketizer::new(CompressionType::Uncompressed),
+            )),
+            ("application", "vnd.onvif.metadata.gzip") => DepacketizerInner::Onvif(Box::new(
+                onvif::Depacketizer::new(CompressionType::GzipCompressed),
+            )),
+            ("application", "vnd.onvif.metadata.exi.onvif") => DepacketizerInner::Onvif(Box::new(
+                onvif::Depacketizer::new(CompressionType::ExiDefault),
+            )),
+            ("application", "vnd.onvif.metadata.exi.ext") => DepacketizerInner::Onvif(Box::new(
+                onvif::Depacketizer::new(CompressionType::ExiInBand),
+            )),
             (_, _) => {
                 log::info!(
                     "no depacketizer for media/encoding_name {}/{}",
                     media,
                     encoding_name
                 );
-                bail!(
+                return Err(format!(
                     "no depacketizer for media/encoding_name {}/{}",
-                    media,
-                    encoding_name
-                );
+                    media, encoding_name
+                ));
             }
         }))
     }
 
-    pub fn parameters(&self) -> Option<&Parameters> {
+    pub fn parameters(&self) -> Option<Parameters> {
         match &self.0 {
             DepacketizerInner::Aac(d) => d.parameters(),
             DepacketizerInner::G723(d) => d.parameters(),
@@ -401,7 +401,7 @@ impl Depacketizer {
         }
     }
 
-    pub fn push(&mut self, input: rtp::Packet) -> Result<(), Error> {
+    pub fn push(&mut self, input: rtp::Packet) -> Result<(), String> {
         match &mut self.0 {
             DepacketizerInner::Aac(d) => d.push(input),
             DepacketizerInner::G723(d) => d.push(input),
@@ -411,13 +411,63 @@ impl Depacketizer {
         }
     }
 
-    pub fn pull(&mut self) -> Result<Option<CodecItem>, Error> {
+    pub fn pull(&mut self, conn_ctx: &ConnectionContext) -> Result<Option<CodecItem>, Error> {
         match &mut self.0 {
-            DepacketizerInner::Aac(d) => d.pull(),
-            DepacketizerInner::G723(d) => d.pull(),
-            DepacketizerInner::H264(d) => d.pull(),
-            DepacketizerInner::Onvif(d) => d.pull(),
-            DepacketizerInner::SimpleAudio(d) => d.pull(),
+            DepacketizerInner::Aac(d) => d.pull(conn_ctx),
+            DepacketizerInner::G723(d) => Ok(d.pull()),
+            DepacketizerInner::H264(d) => Ok(d.pull()),
+            DepacketizerInner::Onvif(d) => Ok(d.pull()),
+            DepacketizerInner::SimpleAudio(d) => Ok(d.pull()),
+        }
+    }
+}
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    // See with: cargo test -- --nocapture codec::tests::print_sizes
+    #[test]
+    fn print_sizes() {
+        for (name, size) in [
+            ("Depacketizer", std::mem::size_of::<Depacketizer>()),
+            (
+                "aac::Depacketizer",
+                std::mem::size_of::<aac::Depacketizer>(),
+            ),
+            (
+                "g723::Depacketizer",
+                std::mem::size_of::<g723::Depacketizer>(),
+            ),
+            (
+                "h264::Depacketizer",
+                std::mem::size_of::<h264::Depacketizer>(),
+            ),
+            (
+                "onvif::Depacketizer",
+                std::mem::size_of::<onvif::Depacketizer>(),
+            ),
+            (
+                "simple_audio::Depacketizer",
+                std::mem::size_of::<simple_audio::Depacketizer>(),
+            ),
+            ("CodecItem", std::mem::size_of::<CodecItem>()),
+            ("VideoFrame", std::mem::size_of::<VideoFrame>()),
+            ("AudioFrame", std::mem::size_of::<AudioFrame>()),
+            ("MessageFrame", std::mem::size_of::<MessageFrame>()),
+            (
+                "SenderReport",
+                std::mem::size_of::<crate::client::rtp::SenderReport>(),
+            ),
+            ("Parameters", std::mem::size_of::<Parameters>()),
+            ("VideoParameters", std::mem::size_of::<VideoParameters>()),
+            ("AudioParameters", std::mem::size_of::<AudioParameters>()),
+            (
+                "MessageParameters",
+                std::mem::size_of::<MessageParameters>(),
+            ),
+        ] {
+            println!("{:-40} {:4}", name, size);
         }
     }
 }
